#!/usr/bin/perl -I/usr/sausalito/perl
# $Id: check_backups.pl, v2.0.0-4 Sat 12 Nov 2011 04:53:40 AM CET mstauber Exp $
# Copyright 2006-2011 Stauber Multimedia Design. All rights reserved.

# This handler / constructor checks for the presence of local and remote backup
# files. It then updates CODB with that information and if need be creates new
# CODB 'Backup' objects..
#
# Additionally it walks through the existing 'Backup' CODB objects and removes
# all CODB 'Backup' objects for which we no longer have local or remote payloads.
#
# Lastly it makes sure that we don't have more remote tarballs than we have set
# 'maxfiles' to. If it discoveres more remote tarballs than we want, it deletes
# the oldes ones and removes the respective 'Backup' object.

$whatami = "constructor";

use CCE;
my $cce = new CCE;
use POSIX qw(isalpha);
use Socket;
use Sys::Hostname;
use File::Spec;
use List::MoreUtils qw(uniq);
use Net::SFTP::Foreign;
use File::Basename;
use POSIX qw(strftime);
use DateTime;

# Helper to redirect SSH errors to /dev/null:
open $ssherr, '>', '/dev/null'; # or die "Unable to open /dev/null";

if ($whatami eq "handler") {
    $cce->connectfd();
}
else {
    $cce->connectuds();
}

# If this is a secondary in a cluster, we can save us the bother:
if (!-d "/vz/private") {
	$cce->bye('SUCCESS');
	exit 0;
}

# Get settings out of CODB to determine the target of the remote backup (if any):
@oids = $cce->find('VserverMasterNode');
($ok, $solbackupValues) = $cce->get($oids[0]);
if ($ok) {
    # Define values:
    $vzdump_enabled		= $solbackupValues->{'vzdump_enabled'};
    $rsync_enabled		= $solbackupValues->{'rsync_enabled'};
    $MN_RemoteBackupEnabled	= $solbackupValues->{'MN_RemoteBackupEnabled'};
    $MN_LocalBackupEnabled	= $solbackupValues->{'MN_LocalBackupEnabled'};
    $MN_BackupLocalLocation	= $solbackupValues->{'MN_BackupLocalLocation'};
    $targethost                 = $solbackupValues->{'MN_BackupRemoteTargetHost'};
    $targetip                   = $solbackupValues->{'MN_BackupRemoteTargetIP'};
    $path                       = $solbackupValues->{'MN_BackupRemoteLocation'};
    $user                       = $solbackupValues->{'MN_BackupRemoteTargetUser'};
    $pass                       = $solbackupValues->{'MN_BackupRemoteTargetUserPass'};
    $port                       = $solbackupValues->{'MN_BackupRemoteTargetPort'};
    $limit                      = $solbackupValues->{'RsyncSpeedLimit'};
    $my_maxfiles                = $solbackupValues->{'vzdump_maxfiles'};

    # Assemble paths:
    $local_location_vzdump = $solbackupValues->{'MN_BackupLocalLocation'} . "/vzbackup";
    $local_location_vzdump =~ s|//|/|g;
    $local_location_vzdump =~ s|/$||g;

    $remote_location_vzdump = $solbackupValues->{'MN_BackupRemoteLocation'};
    $remote_location_vzdump =~ s|//|/|g;
    $remote_location_vzdump =~ s|/$||g;

    $local_location_rsync = $solbackupValues->{'MN_BackupLocalLocation'} . "/rsync";
    $local_location_rsync =~ s|//|/|g;
    $local_location_rsync =~ s|/$||g;

    $remote_location_rsync = $solbackupValues->{'MN_BackupRemoteLocation'};
    $remote_location_rsync =~ s|//|/|g; 
    $remote_location_rsync =~ s|/$||g; 
}

# Check if there is anything to do. Are there any local or remote backups at these
# specified locations?

# Local VZdumps:
if (-d $local_location_vzdump) {
    #print "Local VZdump backup directory found.\n";
    $path = $local_location_vzdump . '/*.tar';
    $exec = `ls -k1 $path|wc -l`;
    if ($exec != "0") {
	#print "Local VZdump tarballs found. \n";

	opendir(DIR, $local_location_vzdump) or die $!;
	while (my $file = readdir(DIR)) {
	    # Use a regular expression to ignore files beginning with a period
    	    next if ($file =~ m/^\./);
    	    if ($file =~ m/^vzdump-(.*)\.tar$/) {
		# Get virt-type, VPSID, date and time parsed from tarball filename:
		@tarball_info = split(/-/, $file);
		$virt_type = $tarball_info[1];
		$vpsid = $tarball_info[2];
		$date = $tarball_info[3];
		$date =~ s/\_/\-/g;
		@timeh = split('\.', $tarball_info[4]);
		$time = $timeh[0];
		$time =~ s/\_/\:/g;
		$datetime = $date . " " . $time;
		$tarfile = $local_location_vzdump . "/" . $file;

		# Check if we already have a CODB object for this tarball:
		$path_helper = "$local_location_vzdump/" . $file;
		@my_local_backups = $cce->findx('Backup', {'vpsid' => $vpsid, 'location' => 'local', 'storage' => $path_helper, 'origin' => 'vzdump'});
		if ($my_local_backups[0] eq "") {
		    $proc_helper = `ps axf|grep /usr/sbin/vzdump |grep -v grep|wc -l`;
		    if ($proc_helper > '0') {
			# No CODB object for this tarball present, but VZdump is currently running. Laying off creation of CODB objects.
		    }
		    else {
			# VZdump is not running. Creating a matching CODB object for this tarball:
		        ($ok) = $cce->create('Backup', {
		            'origin' => 'vzdump',
		            'vpsid' => $vpsid,
		            'virttype' => $virt_type,
		            'location' => 'local',
		            'LBtimestamp' => $datetime,
		            'storage' => $tarfile,
		            'timestamp' => time()
		        });
		    }
		}
	    }
	}
	closedir(DIR);
    }
}

# Local RSYNCs:
if (-d $local_location_rsync) {
    # Local RSYNC backup directory found.
    $exec = `ls -d1 $local_location_rsync|wc -l`;
    if ($exec != "0") {
	# Local RSYNC directories found
	opendir(DIR, $local_location_rsync) or die $!;
	while (my $file = readdir(DIR)) {
	    # Use a regular expression to ignore files beginning with a period
    	    next if ($file =~ m/^\./);
    	    next if ($file =~ m/^vzdump-(.*)\.tar$/);
	    if (-f "$local_location_rsync/$file/etc/vzdump/vps.conf") {
		# Found locally stored VPS backup config at $local_location_rsync/$file/etc/vzdump/vps.conf

		$virt_type = 'openvz';
		$vpsid = $file;
		$modify = `stat $local_location_rsync/$file/etc/vzdump/vps.conf | grep ^Modify:`;
		@mod_helper = split('\s+', $modify);
		$date = $mod_helper[1];
		@timeh = split('\.', $mod_helper[2]);
		$time = $timeh[0];
		$datetime = $date . " " . $time;

		# Check if we already have a CODB object for this RSYNC backup:
		$path_helper = $local_location_rsync . "/" . $file;
		@my_local_backups = $cce->findx('Backup', {'vpsid' => $vpsid, 'location' => 'local', 'storage' => $path_helper, 'origin' => 'rsync'});
		if ($my_local_backups[0] eq "") {
		    # VPS $vpsid needs a CODB object:
		    if (-f "/tmp/vz_vserver_backup.lock") {
			# No CODB object for this RSYNC backup present, but RSYNC is currently running. Laying off creation of CODB objects.
		    }
		    else {
			# VZbackup is not running. Creating a matching CODB object for this rsync backup:
		        ($ok) = $cce->create('Backup', {
		            'origin' => 'rsync',
		            'vpsid' => $vpsid,
		            'virttype' => $virt_type,
		            'location' => 'local',
		            'LBtimestamp' => $datetime,
		            'storage' => $path_helper,
		            'timestamp' => time()
		        });
		    }
		}
	    }
	}
	closedir(DIR);
    }
}

# Remote VZdumps:
if ($user && $pass && $port && ($targethost||$targetip) && $remote_location_vzdump) {
    if ($targethost) {
        $target = $user . '@' . $targethost;
    }
    else {
        $target = $user . '@' . $targetip;
    }

    # Connect to remote storage:
    no warnings 'Net::SFTP::Foreign';
    $sftp = Net::SFTP::Foreign->new($target, password => $pass, port => $port, autodie => 0, stderr_fh => $ssherr);
    $sftp->die_on_error("Unable to connect to remote host");

    # Glob for tarballs:
    $num_tarballs = $sftp->glob("$remote_location_vzdump/vzdump-*.tar");

    if ($num_tarballs > '0') {
	# Get tarball filenames:
	@ls = @{ $sftp->ls("$remote_location_vzdump") };
	for $file (@ls) {
	    # Get filenames of valid VZdumps:
    	    if (($file->{filename} ne '.') && ($file->{filename} ne '..') && ($file->{filename} =~ /^vzdump-(.*)\.tar$/)) {

		# Get virt-type, VPSID, date and time parsed from tarball filename:
		@tarball_info = split(/-/, $file->{filename});
		$virt_type = $tarball_info[1];
		$vpsid = $tarball_info[2];
		$date = $tarball_info[3];
		$date =~ s/\_/\-/g;
		@timeh = split('\.', $tarball_info[4]);
		$time = $timeh[0];
		$time =~ s/\_/\:/g;
		$datetime = $date . " " . $time;
		$tarfile = $remote_location_vzdump . "/" . $file->{filename};

		# Check if we already have a CODB object for this tarball:
		$path_helper = "$remote_location_vzdump/" . $file->{filename};
		@my_remote_backups = $cce->findx('Backup', {'vpsid' => $vpsid, 'location' => 'remote', 'storage' => $path_helper, 'origin' => 'vzdump'});
		if ($my_remote_backups[0] eq "") {
		    $proc_helper = `ps axf|grep /usr/sbin/vzdump |grep -v grep|wc -l`;
		    if ($proc_helper > '0') {
			# No CODB object for this tarball present, but VZdump is currently running. Laying off creation of CODB objects.
		    }
		    else {
			# VZdump is not running. Creating a matching CODB object for this tarball:
		        ($ok) = $cce->create('Backup', {
		            'origin' => 'vzdump',
		            'vpsid' => $vpsid,
		            'virttype' => $virt_type,
		            'location' => 'remote',
		            'LBtimestamp' => $datetime,
		            'storage' => $tarfile,
		            'timestamp' => time()
		        });
		    }
		}
    	    }
	}
    }

}

# Remote RSYNCs:
if ($user && $pass && $port && ($targethost||$targetip) && $remote_location_rsync) {
    if ($targethost) {
        $target = $user . '@' . $targethost;
    }
    else {
        $target = $user . '@' . $targetip;
    }

    # Connect to remote storage:
    no warnings 'Net::SFTP::Foreign';
    $sftp = Net::SFTP::Foreign->new($target, password => $pass, port => $port, autodie => 0, stderr_fh => $ssherr);
    $sftp->die_on_error("Unable to connect to remote host");

    # Glob for directories with '/etc/sysconfig/network' in them as any Linux VPS will have them:
    $num_rsyncs = $sftp->glob("$remote_location_rsync/*/etc/sysconfig/network");
    if ($num_rsyncs > '0') {
	# Get VPSID from RSYNC directory name:
	@ls = @{ $sftp->ls("$remote_location_rsync") };
	for $file (@ls) {
	    $path_helper = $remote_location_rsync . "/" . $file->{filename};
	    # Ignore these files:
    	    if (($file->{filename} == '.') || ($file->{filename} == '..') || ($file->{filename} =~ /^vzdump-(.*)\.tar$/)) {
    		next;
    	    }
	    # Be more specific: See if the 'hit' is a directory and contains /etc/vzdump/vps.conf:
    	    $valid_rsync = $sftp->glob("$path_helper/etc/vzdump/vps.conf");
    	    if ($valid_rsync == "1") {
		# OK, it looks like a RSYNC backed up VPS.
		$virt_type = 'openvz';
		$vpsid = $file->{filename};
		# Get date and time when this backup set was last written to by stat'ing vps.conf:
		$modify_helper = $sftp->stat("$path_helper/etc/vzdump/vps.conf");
		# Haggle with DateTime to transform epoch date into human readable format:
		$dt_helper = DateTime->from_epoch( epoch => $modify_helper->mtime );
		$date = $dt_helper->ymd;
		$time = $dt_helper->hms;
		$datetime = $date . " " . $time;

		# Check if we already have a CODB object for this RSYNC backup:
		@my_remote_backups = $cce->findx('Backup', {'vpsid' => $vpsid, 'location' => 'remote', 'storage' => $path_helper, 'origin' => 'rsync'});
		if ($my_remote_backups[0] eq "") {
		    # VPS $vpsid needs a CODB object:
		    if (-f "/tmp/vz_vserver_backup.lock") {
			# No CODB object for this RSYNC backup present, but RSYNC is currently running. Laying off creation of CODB objects.
		    }
		    else {
			# VZbackup is not running. Creating a matching CODB object for this rsync backup:
		        ($ok) = $cce->create('Backup', {
		            'origin' => 'rsync',
		            'vpsid' => $vpsid,
		            'virttype' => $virt_type,
		            'location' => 'remote',
		            'LBtimestamp' => $datetime,
		            'storage' => $path_helper,
		            'timestamp' => time()
		        });
		    }
		}
    	    }
	}
    }
}

# Now work in reverse: Walk through the 'Backup' objects and check if their payload is still present. If not, remove the CODB objects:
$proc_helper = `ps axf|grep /usr/sbin/vzdump |grep -v grep|wc -l`;
if ($proc_helper > '0') {
	# No CODB object for this tarball present, but VZdump is currently running. Laying off deletion of CODB objects.
}
else {
    # Backup currently not running. 

    # Verify local backups:
    @my_local_verify = $cce->findx('Backup', {'location' => 'local'});
    for $entry (@my_local_verify) {
	($ok, $old_entry) = $cce->get($entry);
	# Handle vzdump tarballs:
	if ($old_entry->{'origin'} eq 'vzdump') {
	    $is_file = $old_entry->{'storage'};
	    if (!-f "$is_file") {
    		($ok) = $cce->destroy($entry);
	    }
	}
	# Handle rsync directories:
	if ($old_entry->{'origin'} eq 'rsync') {
	    $is_dir = $old_entry->{'storage'};
	    if (!-d "$is_dir") {
    		($ok) = $cce->destroy($entry);
	    }
	}
    }

    # Verify remote backups:
    if ($user && $pass && $port && ($targethost||$targetip) && $remote_location_vzdump) {
	if ($targethost) {
    	    $target = $user . '@' . $targethost;
	}
	else {   
    	    $target = $user . '@' . $targetip;
	}

	# Connect to remote storage:
	no warnings 'Net::SFTP::Foreign';
	$sftp = Net::SFTP::Foreign->new($target, password => $pass, port => $port, autodie => 0, stderr_fh => $ssherr);
	$sftp->die_on_error("Unable to connect to remote host");

	# Check remote RSYNCs:
	@my_local_verify = $cce->findx('Backup', {'location' => 'remote', 'origin' => 'rsync'});
	for $entry (@my_local_verify) {
	    ($ok, $old_entry) = $cce->get($entry);
	    # Handle rsync directories:
    	    # Is this a directory with the vzdump config directory in /etc/?
    	    $fh1 = $sftp->stat("$old_entry->{'storage'}/etc/vzdump/vps.conf");
    	    unless ($fh1) {
	        ($ok) = $cce->destroy($entry);
	    }
	}

	# Check remote VZDUMPs:
	@my_local_verify = $cce->findx('Backup', {'location' => 'remote', 'origin' => 'vzdump'});
	for $entry (@my_local_verify) {
	    ($ok, $old_entry) = $cce->get($entry);
	    # Handle vzdump tarballs:
	    # Glob for tarball:
	    $num_tarballs = $sftp->glob("$old_entry->{'storage'}");
	    if ($num_tarballs lt '1') {
	        ($ok) = $cce->destroy($entry);
	    }
	}
    }
}

# Remove remote CODB 'Backup' tarballs if we have more 'remote' Backup objects than 
# --maxfiles allows. First build an array with all 'remote' 'Backup' objects of this
# VPSID:

@vpsids_remote_backups_all = $cce->findx('Backup', {'location' => 'remote', 'origin' => 'vzdump'});
for $finding (@vpsids_remote_backups_all) {
    ($ok, $back_obj) = $cce->get($finding);
    $vmid = $back_obj->{'vpsid'};
    push(@all_vpsids, $vmid);
}

@my_vpsids = uniq(@all_vpsids);

for $vmid (@my_vpsids) {
    $num_of_remotes = "0";
    @my_remote_backups = $cce->findx('Backup', {'vpsid' => $vmid, 'location' => 'remote', 'origin' => 'vzdump'});
    for $entry (@my_remote_backups) {
        $num_of_remotes++;
    }

    # Remove excess remote backups:
    @sorted = ();
    @remote_stored = ();
    if ($num_of_remotes > $my_maxfiles) {
        # We have more remote backups than we want to keep:
        for $xentry (@my_remote_backups) {
                ($ok, $old_entry) = $cce->get($xentry);
                push(@remote_stored, $old_entry->{'timestamp'});
        }
        @sorted = sort { $a <=> $b } @remote_stored;

        # Remove as many newest entries from @sorted as we have maxfiles set to:
        $z = $my_maxfiles;
        for ($z = $my_maxfiles; $z > 0; $z--) {
            pop(@sorted);
        }

        # We now have a sorted array with the remote backups. Remove the oldest tarball(s):
        foreach $goner (@sorted) {
            @my_single = $cce->find('Backup', {'vpsid' => $vmid, 'timestamp' => $goner, 'location' => 'remote', 'origin' => 'vzdump'});
            ($ok, $expire_goner) = $cce->get($my_single[0]);
            $exp_path = $expire_goner->{'storage'};

            # Assemble target login details:
            if ($targethost) {
                $target = $user . '@' . $targethost;
            }
            else {
                $target = $user . '@' . $targetip;
            }

            # Connect to remote storage:
            no warnings 'Net::SFTP::Foreign';
            $sftp = Net::SFTP::Foreign->new($target, password => $pass, port => $port, autodie => 0, stderr_fh => $ssherr);
            $sftp->die_on_error("Unable to connect to remote host");

            # Does the old file exist?
            $fh1 = $sftp->stat("$exp_path");
            if ($fh1) {
                # Remove file:
                $fh2 = $sftp->remove("$exp_path");
                if ($fh2) {
                    # Remove CODB Object:
                    ($ok) = $cce->destroy($my_single[0]);
                }
            }
        }
    }
}

$cce->bye('SUCCESS');
exit 0;

